#define _GNU_SOURCE

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <libgen.h>

#ifndef RENAME_EXCHANGE
#define RENAME_EXCHANGE (1<<1)
#endif

#define UTILITY_NAME "alterator_users_helper"
#define BUFSIZE 256
#define NUM_OF_REQUIRED_PARAMS 2

void printHelp() {
    printf("Usage:\n"
            " %s [options] file...\n\n"
            "Options:\n"
            " -f, --from\t old path to the file (directory)\n"
            " -t, --to\t new path to the file (directory)\n"
            " -u, --uid\t set UID for current utility\n\n"
            " -h, --help\t display this help\n", UTILITY_NAME);
}

int renameFile(const char* from, char* to) {
    struct stat stat_buf;

    if (-1 == lstat(to, &stat_buf)) return -1;
    if ( !(S_ISREG(stat_buf.st_mode) || S_ISLNK(stat_buf.st_mode)) ) {
        fprintf(stderr, "%s - is neither a file nor a symbolic link.\n", to);
        return -1;
    }

    int fd = open(to, O_WRONLY);
    if (-1 != fd) {
        close(fd);
        return renameat2(AT_FDCWD, from, AT_FDCWD, to, RENAME_EXCHANGE);
    }

    fd = open(to, O_WRONLY|O_CREAT, 0644);
    if (-1 != fd) close(fd);
    else {
        fprintf(stderr, "Failed to create file %s\n", basename(to));
        return -1;
    }

    return renameat2(AT_FDCWD, from, AT_FDCWD, to, RENAME_EXCHANGE);
}

int main(int argc, char* argv[]) {
    struct option opts[] = {
        {"help", no_argument, 0, 'h'},
        {"from", required_argument, 0, 'f'},
        {"to", required_argument, 0, 't'},
        {"uid", required_argument, 0, 'u'}
    };

    char path_from[BUFSIZE] = "";
    char path_to[BUFSIZE] = "";

    int opchar = 0;
    unsigned int data_cnt = 0;
    while (-1 != (opchar = getopt_long(argc, argv, "W;hW;f:W;t:W;u:", opts, NULL))) {
        switch (opchar) {
        case 'h':
            printHelp();
            return EXIT_SUCCESS;
        case 'f':
            if (data_cnt != NUM_OF_REQUIRED_PARAMS) {
                strcpy(path_from, optarg);
                ++data_cnt;
            }
            break;
        case 't':
            if (data_cnt != NUM_OF_REQUIRED_PARAMS) {
                strcpy(path_to, optarg);
                ++data_cnt;
            }
            break;
        case 'u':
            if (data_cnt != NUM_OF_REQUIRED_PARAMS) {
                if (-1 == setuid(atoi(optarg))) {
                    fprintf(stderr, "Failed to set UID for current utility");
                    return -1;
                }
                ++data_cnt;
            }
            break;
        default:
            fprintf(stderr, "Invalid parameters passed!\n");
        }
    }

    if (opterr) {
        if (data_cnt < NUM_OF_REQUIRED_PARAMS) {
            fprintf(stderr, "%s: not enough arguments\n"
                    "Try '%s --help' for more information.\n",
                    UTILITY_NAME, UTILITY_NAME);
        }
        else {
            if (-1 == renameFile(path_from, path_to))
                fprintf(stderr, "Failed to change file name or location.\n");
        }
    }

    return EXIT_SUCCESS;
}
